from orchestrator.workflows import LazyWorkflowInstance

LazyWorkflowInstance("workflows.example2.create_example2", "create_example2")
LazyWorkflowInstance("workflows.example2.modify_example2", "modify_example2")
LazyWorkflowInstance("workflows.example2.terminate_example2", "terminate_example2")
LazyWorkflowInstance("workflows.example2.validate_example2", "validate_example2")
LazyWorkflowInstance("workflows.example1.create_example1", "create_example1")
LazyWorkflowInstance("workflows.example1.modify_example1", "modify_example1")
LazyWorkflowInstance("workflows.example1.terminate_example1", "terminate_example1")
LazyWorkflowInstance("workflows.example1.validate_example1", "validate_example1")
LazyWorkflowInstance("workflows.example4.create_example4", "create_example4")
LazyWorkflowInstance("workflows.example4.modify_example4", "modify_example4")
LazyWorkflowInstance("workflows.example4.terminate_example4", "terminate_example4")
LazyWorkflowInstance("workflows.example4.validate_example4", "validate_example4")
